@@ -30,7 +30,7 @@ class Socket(object):
             packets = [self.queue.get(timeout=self.server.ping_timeout)]
             self.queue.task_done()
         except self.server._async['queue'].Empty:
-            raise IOError()
+            raise exceptions.QueueEmpty()
         if packets == [None]:
             return []
         try:
@@ -89,7 +89,7 @@ class Socket(object):
                                                           start_response)
         try:
             packets = self.poll()
-        except IOError as e:
+        except exceptions.QueueEmpty as e:
             self.close(wait=False)
             raise e
         return packets
@@ -109,12 +109,18 @@ class Socket(object):
         """Close the socket connection."""
         if not self.closed and not self.closing:
             self.closing = True
-            self.server._trigger_event('disconnect', self.sid, async=False)
+            reraise_exc = None
+            try:
+                self.server._trigger_event('disconnect', self.sid, async=False)
+            except Exception as e:
+                reraise_exc = e
             if not abort:
                 self.send(packet.Packet(packet.CLOSE))
             self.closed = True
             if wait:
                 self.queue.join()
+            if reraise_exc:
+                raise reraise_exc
 
     def _upgrade_websocket(self, environ, start_response):
         """Upgrade the connection from polling to websocket."""
@@ -166,7 +172,7 @@ class Socket(object):
             while True:
                 try:
                     packets = self.poll()
-                except IOError:
+                except exceptions.QueueEmpty:
                     break
                 if not packets:
                     # empty packet list returned -> connection closed
@@ -181,6 +187,7 @@ class Socket(object):
         self.server.logger.info(
             '%s: Upgrade to websocket successful', self.sid)
 
+        reraise_exc = None
         while True:
             p = None
             try:
@@ -197,9 +204,17 @@ class Socket(object):
                 self.receive(pkt)
             except exceptions.UnknownPacketError:
                 pass
+            except Exception as e:
+                # if we get an unexpected exception (such as something in an
+                # application event handler) we close the connection properly
+                # and then reraise the exception
+                reraise_exc = e
+                break
 
         self.queue.put(None)  # unlock the writer task so that it can exit
         writer_task.join()
         self.close(wait=True, abort=True)
+        if reraise_exc:
+            raise reraise_exc
 
         return []
